use std::collections::HashMap;
use std::iter::Peekable;

use roxmltree::{Children, Node};

use crate::law::LawBuilder;
use crate::paragraph::parser::liste::Liste;
use crate::paragraph::parser::{Absatz, AbsatzAbs, Content, Fzinhalt, Kzinhalt, Ueberschrift};

use super::table::Table;

#[derive(Debug, PartialEq, Default)]
pub(crate) struct Abschnitt {
    metadata: HashMap<String, String>,
    pub(crate) cont: bool,
}

impl Abschnitt {
    pub(crate) fn parse(n: Node, builder: &mut LawBuilder) -> Abschnitt {
        assert!(n.tag_name().name() == "abschnitt");

        let mut ret = Abschnitt::default();
        let mut c = n.children().peekable();

        Self::skip_static_fields(&mut c);

        ret.handle_metadata(&mut c, builder);

        if !ret.handle_headers(&mut c, builder) {
            return ret;
        }

        let mut absatze = Vec::new();

        // Special handling of first paragraph (needs id)...
        let (par_id, first_abs) = ret.parse_absatz(&mut c);
        let par_id = match par_id {
            Some(par_id) => par_id,
            None => panic!("First paragraph needs to have an id, not found"),
        };
        absatze.push(first_abs);

        // ... and then there can be as many 'Absätze' as our law-setter wants
        while let Some(child) = c.peek() {
            if AbsatzAbs::test(child) {
                let (_, absatz) = ret.parse_absatz(&mut c);
                absatze.push(absatz);
            } else {
                break;
            }
        }

        if absatze.len() == 1 {
            builder.new_par(par_id, absatze[0].clone());
        } else {
            let mut contents = Vec::new();
            for a in &absatze {
                contents.push(a.clone());
            }
            builder.new_par(par_id, Content::Item(contents));
        }

        // Skip all UeberschriftTitle and Absatz
        while let Some(child) = c.peek() {
            if Ueberschrift::test(child, "titel") {
                c.next();
                continue;
            }
            if Absatz::test_with_typ(child, "erltext") {
                c.next();
                continue;
            }
            break;
        }

        // assert_eq!(c.next(), None);

        ret.cont = true;
        ret
    }

    // There are paragraph-specific meta-data at the top of each xml file. We parse those. When we
    // encounter the title "Text" the real content starts, we stop parsing meta data.
    fn handle_metadata(&mut self, c: &mut Peekable<Children>, builder: &mut LawBuilder) {
        loop {
            let key = Ueberschrift::parse(c.next().unwrap(), "titel").content;

            // We are done with meta-data parsing
            if key == "Text" {
                break;
            }

            let absatz = Absatz::parse(
                c.next()
                    .expect("Expected absatz after title in par headers"),
            );

            let value = absatz.content;

            // We want ot use this information in our markdown output.
            // TODO: Use all metadata, instead of this specific call
            if key == "Beachte" {
                builder.add_next_para_note(value.clone());
            }

            self.metadata.insert(key, value);
            continue;
        }
    }

    // At the beginning of each 'Abschnitt' there are 4 static fields. Since they don't provide any
    // value, we skip them with this function.
    // If they are not there, we panic (unwrap), as we should take a look why they changed that.
    fn skip_static_fields(node: &mut Peekable<Children>) {
        Kzinhalt::parse(node.next().unwrap()); // "Bundesrecht konsolidiert"
        Kzinhalt::parse(node.next().unwrap()); // "Bundesrecht konsolidiert"
        Fzinhalt::parse(node.next().unwrap()); // "www.ris.bka.gv.at" and "Seite X von Y"
        Fzinhalt::parse(node.next().unwrap()); // "www.ris.bka.gv.at" and "Seite X von Y"
    }

    // we have optionally headers. Such as "Einleitung", "Von den bürgerlichen Gesetzen üerhaupt,"
    // etc. If we have headers which indicate that we are done and we want to stop parsing
    // ("anlage" + "Artikel" we indicate this wish by returning false.
    fn handle_headers(&self, c: &mut Peekable<Children>, builder: &mut LawBuilder) -> bool {
        while let Some(child) = c.peek() {
            // Schiffahrtsgesetz: stop @ anlagen (for now)
            if Ueberschrift::test(child, "anlage") {
                return false;
            }
            if Ueberschrift::test(child, "g1") {
                let ueberschrift = Ueberschrift::parse(c.next().unwrap(), "g1");
                if ueberschrift.content.trim().starts_with("Artikel") {
                    return false;
                }
                builder.new_header(&ueberschrift.content);
            } else if Ueberschrift::test(child, "g2") {
                let ueberschrift = Ueberschrift::parse(c.next().unwrap(), "g2");
                builder.new_desc(&ueberschrift.content);
            } else if Ueberschrift::test(child, "g1min") {
                let ueberschrift = Ueberschrift::parse(c.next().unwrap(), "g1min");
                builder.new_header(&ueberschrift.content);
            } else if Ueberschrift::test(child, "art") {
                let ueberschrift = Ueberschrift::parse(c.next().unwrap(), "art");
                if ueberschrift.content.trim().starts_with("Artikel") {
                    return false;
                }
            } else {
                break;
            }
        }

        // while (not if) because we can have two subsequent paraheaders (e.g. § 405 abgb)
        while let Some(child) = c.peek() {
            if Ueberschrift::test(child, "para") {
                builder
                    .new_next_para_header(&Ueberschrift::parse(c.next().unwrap(), "para").content);
                continue;
            }
            break;
        }
        true
    }

    // Parses one logical 'Absatz'. If there's a List or Table after the Absatz, RIS assumes this
    // one to be included in the paragraph
    //
    // # Returns
    // - String: (optional) paragraph id
    // - Content: content of the paragraph
    fn parse_absatz(&self, c: &mut Peekable<Children>) -> (Option<String>, Content) {
        let absatz = AbsatzAbs::parse(c.next().unwrap());
        let par_id = absatz.gldsym;

        let mut content = Vec::new();
        content.push(Content::Text(absatz.content));

        // If there's a  "liste" after an "absatz", the "liste" should be part of the "absatz"
        while let Some(child) = c.peek() {
            if Liste::test(child) {
                content.push(Liste::parse_full(c).get_content())
            } else if Table::test(child) {
                // If there's a  "table" after an "absatz", the "table" should be part of the "absatz"
                let table = Table::parse_full(c);
                content.extend(table.iter().cloned());
            } else if Absatz::test_with_typ(child, "satz") {
                // After a 'absatz' there can be a '<absatz typ="satz"' which should be part of the first absatz
                // (e.g. 1209 ABGB)
                content.push(Content::Text(Absatz::parse(c.next().unwrap()).content))
            } else if Absatz::test_with_typ(child, "erltext") {
                // After a 'absatz' there can be a '<absatz typ="erltext"' which should be part of the first absatz
                // (e.g. 51 StGB)
                content.push(Content::Text(Absatz::parse(c.next().unwrap()).content))
            } else {
                break;
            }
        }
        if content.len() == 1 {
            (par_id, content[0].clone())
        } else {
            (par_id, Content::List(content))
        }
    }
}
